import numpy as np
import open3d as o3d
import scipy.constants

from ..base import scattering_t
from ..electromagnetics import empropagation as EM
from ..geometry import targets as TL
from ..raycasting import rayfunctions as RF


def calculate_scattering(aperture_coords,
                         sink_coords,
                         excitation_function,
                         antenna_solid,
                         desired_E_axis,
                         scatter_points=None,
                         wavelength=1.0,
                         scattering=0,
                         elements=False,
                         sampling_freq=1e9,
                         num_samples=10000,
                         mesh_resolution=0.5):
    """
    Based upon the parameters given, calculate the time domain scattering for the apertures and sinks.
    The defaults for sampling time are based upon a sampling rate of 1GHz, and a sample time of 1ms.

    Parameters
    ----------
    aperture_coords : open3d point cloud
        source coordinates
    sink_coords : open3d point cloud
        sink coordinates
    antenna_solid : structure class
        the class should contain all the environment for scattering, providing the blocking for the rays
    desired_E_axis : 1D numpy array of floats
        the desired excitation vector, can be a 1*3 array or a n*3 array if multiple different exciations are desired in one lauch
    scatter_points : open3d point cloud
        the scattering points in the environment. Defaults to [None], in which case scattering points will be generated from the antenna_solid. If no scattering should be considered then set scattering to [0].
    wavelength : float
        the wavelength of interest in metres
    scattering : int
        the number of reflections to be considered, defaults to [0], but up to 2 can be considered. The higher this number to greater to computational effort, and for most situations 1 should be ample.
    elements : boolean
        whether the sources and sinks should be considered as elements of a phased array, or a fixed phase aperture like a horn or reflector
    sampling_freq : float
        the desired sampling frequency, should be twice the highest frequency of interest to avoid undersampling artifacts
    num_samples : int
        the length of the desired sampling, can be calculated from the desired model time
    mesh_resolution : float
        the desired mesh resolution in terms of wavelengths if scattering points are not provided. A scattering mesh is generated on the surfaces of all provided trianglemesh structures.

    Return
    -------
    Ex : numpy array of float
        the x directed voltage at the sink coordinates in the time domain, if elements=True, then it will be an array of num_sinks * num_samples, otherwise it will be a 1D array
    Ey : numpy array of float
        the y directed voltage at the sink coordinates in the time domain, if elements=True, then it will be an array of num_sinks * num_samples, otherwise it will be a 1D array
    Ez : numpy array of float
        the z directed voltage at the sink coordinates in the time domain, if elements=True, then it will be an array of num_sinks * num_samples, otherwise it will be a 1D array
    Waketimes : numpy array of float
        the shortest time required for a ray to reach any sink from any source, as long as elements=False. If elements=True, then this is not implemented in the same way, and will return the shortest time required for the final source.
    """



    time_index = np.linspace(0, num_samples/sampling_freq, num_samples)
    if desired_E_axis.size > 3:
        # multiple excitations requried
        multiE = True
    else:
        multiE = False

    num_sources = len(np.asarray(aperture_coords.points))
    num_sinks = len(np.asarray(sink_coords.points))
    environment_triangles = antenna_solid.triangles_base_raycaster()
    if scattering == 0:
        # only use the aperture point cloud, no scattering required.
        scatter_points = o3d.geometry.PointCloud()

        if ~multiE:
            conformal_E_vectors = EM.calculate_conformalVectors(desired_E_axis,
                                                                np.asarray(aperture_coords.normals).astype(np.float32))
        else:
            conformal_E_vectors = EM.calculate_conformalVectors(desired_E_axis[0, :],
                                                                np.asarray(aperture_coords.normals).astype(np.float32))

        unified_model = np.append(np.asarray(aperture_coords.points).astype(np.float32),
                                  np.asarray(sink_coords.points).astype(np.float32), axis=0)
        unified_normals = np.append(np.asarray(aperture_coords.normals).astype(np.float32),
                                    np.asarray(sink_coords.normals).astype(np.float32), axis=0)
        unified_weights = np.ones((unified_model.shape[0], 3), dtype=np.complex64)
        unified_weights[0:num_sources,
        :] = conformal_E_vectors / num_sources  # set total amplitude to 1 for the aperture
        unified_weights[num_sources:num_sources + num_sinks,
        :] = 1 / num_sinks  # set total amplitude to 1 for the aperture
        point_informationv2 = np.empty((len(unified_model)), dtype=scattering_t)
        # set all sources as magnetic current sources, and permittivity and permeability as free space
        point_informationv2[:]['Electric'] = True
        point_informationv2[:]['permittivity'] = 8.8541878176e-12
        point_informationv2[:]['permeability'] = 1.25663706212e-6
        # set position, velocity, normal, and weight of sources
        point_informationv2[0:num_sources]['px'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['py'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['pz'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 2]
        point_informationv2[0:num_sources]['nx'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['ny'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['nz'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 2]
        # set position and velocity of sinks
        point_informationv2[num_sources:(num_sources + num_sinks)]['px'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['py'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['pz'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 2]
        point_informationv2[num_sources:(num_sources + num_sinks)]['nx'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['ny'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['nz'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 2]

        point_informationv2[:]['ex'] = unified_weights[:, 0]
        point_informationv2[:]['ey'] = unified_weights[:, 1]
        point_informationv2[:]['ez'] = unified_weights[:, 2]
        scatter_mask = np.zeros((point_informationv2.shape[0]), dtype=np.int32)
        scatter_mask[0:num_sources] = 0
        scatter_mask[(num_sources + num_sinks):] = 0

    else:
        # create scatter points on antenna solids based upon a half wavelength square
        if scatter_points is None:
            scatter_points, areas = TL.source_cloud_from_shape(antenna_solid, 1e-6, (wavelength * mesh_resolution) ** 2)

        if ~multiE:
            conformal_E_vectors = EM.calculate_conformalVectors(desired_E_axis,
                                                                np.asarray(aperture_coords.normals).astype(np.float32))
        else:
            conformal_E_vectors = EM.calculate_conformalVectors(desired_E_axis[0, :],
                                                                np.asarray(aperture_coords.normals).astype(np.float32))

        unified_model = np.append(np.append(np.asarray(aperture_coords.points).astype(np.float32),
                                            np.asarray(sink_coords.points).astype(np.float32), axis=0),
                                  np.asarray(scatter_points.points).astype(np.float32), axis=0)
        unified_normals = np.append(np.append(np.asarray(aperture_coords.normals).astype(np.float32),
                                              np.asarray(sink_coords.normals).astype(np.float32), axis=0),
                                    np.asarray(scatter_points.normals).astype(np.float32), axis=0)
        unified_weights = np.ones((unified_model.shape[0], 3), dtype=np.complex64)
        unified_weights[0:num_sources,
        :] = conformal_E_vectors / num_sources  # set total amplitude to 1 for the aperture
        unified_weights[num_sources:num_sources + num_sinks,
        :] = 1 / num_sinks  # set total amplitude to 1 for the aperture
        unified_weights[num_sources + num_sinks:, :] = 1 / len(
            np.asarray(scatter_points.points))  # set total amplitude to 1 for the aperture
        point_informationv2 = np.empty((len(unified_model)), dtype=scattering_t)
        # set all sources as magnetic current sources, and permittivity and permeability as free space
        point_informationv2[:]['Electric'] = True
        point_informationv2[:]['permittivity'] = 8.8541878176e-12
        point_informationv2[:]['permeability'] = 1.25663706212e-6
        # set position, velocity, normal, and weight of sources
        point_informationv2[0:num_sources]['px'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['py'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['pz'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 2]
        point_informationv2[0:num_sources]['nx'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['ny'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['nz'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 2]
        # point_informationv2[0:num_sources]['ex']=unified_weights[0:num_sources,0]
        # point_informationv2[0:num_sources]['ey']=unified_weights[0:num_sources,1]
        # point_informationv2[0:num_sources]['ez']=unified_weights[0:num_sources,2]
        # set position and velocity of sinks
        point_informationv2[num_sources:(num_sources + num_sinks)]['px'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['py'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['pz'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 2]
        # point_informationv2[num_sources:(num_sources+num_sinks)]['vx']=0.0
        # point_informationv2[num_sources:(num_sources+num_sinks)]['vy']=0.0
        # point_informationv2[num_sources:(num_sources+num_sinks)]['vz']=0.0
        point_informationv2[num_sources:(num_sources + num_sinks)]['nx'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['ny'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['nz'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 2]
        point_informationv2[(num_sources + num_sinks):]['px'] = np.asarray(scatter_points.points).astype(np.float32)[:,
                                                                0]
        point_informationv2[(num_sources + num_sinks):]['py'] = np.asarray(scatter_points.points).astype(np.float32)[:,
                                                                1]
        point_informationv2[(num_sources + num_sinks):]['pz'] = np.asarray(scatter_points.points).astype(np.float32)[:,
                                                                2]
        point_informationv2[(num_sources + num_sinks):]['nx'] = np.asarray(scatter_points.normals).astype(np.float32)[:,
                                                                0]
        point_informationv2[(num_sources + num_sinks):]['ny'] = np.asarray(scatter_points.normals).astype(np.float32)[:,
                                                                1]
        point_informationv2[(num_sources + num_sinks):]['nz'] = np.asarray(scatter_points.normals).astype(np.float32)[:,
                                                                2]
        point_informationv2[:]['ex'] = unified_weights[:, 0]
        point_informationv2[:]['ey'] = unified_weights[:, 1]
        point_informationv2[:]['ez'] = unified_weights[:, 2]
        #scatter_mask = np.zeros((point_informationv2.shape[0]), dtype=np.int32)
        #scatter_mask[0:num_sources] = 0
        #scatter_mask[(num_sources + num_sinks):] = scattering

    # full_index, initial_index = RF.integratedraycastersetup(num_sources,
    #                                                        num_sinks,
    #                                                        point_informationv2,
    #                                                        RF.convertTriangles(antenna_solid),
    #                                                        scatter_mask)
    full_index, rays = RF.workchunkingv2(np.asarray(aperture_coords.points).astype(np.float32),
                                         np.asarray(sink_coords.points).astype(np.float32),
                                         np.asarray(scatter_points.points).astype(np.float32),
                                         environment_triangles, scattering + 1)

    if not elements:
        # create efiles for model
        if multiE:
            Ex = np.zeros((desired_E_axis.shape[1],num_samples), dtype=np.float64)
            Ey = np.zeros((desired_E_axis.shape[1],num_samples), dtype=np.float64)
            Ez = np.zeros((desired_E_axis.shape[1],num_samples), dtype=np.float64)
            for e_inc in range(desired_E_axis.shape[0]):
                conformal_E_vectors = EM.calculate_conformalVectors(desired_E_axis[e_inc, :],
                                                                    np.asarray(aperture_coords.normals).astype(
                                                                        np.float32))
                unified_weights[0:num_sources, :] = conformal_E_vectors / num_sources
                point_informationv2[:]['ex'] = unified_weights[:, 0]
                point_informationv2[:]['ey'] = unified_weights[:, 1]
                point_informationv2[:]['ez'] = unified_weights[:, 2]
                scattering_coefficient = (1 / 4 * scipy.constants.pi)
                TimeMap, WakeTimes = EM.TimeDomainv3(num_sources,
                                                     num_sinks,
                                                     point_informationv2,
                                                     full_index,
                                                     scattering_coefficient,
                                                     wavelength,
                                                     excitation_function,
                                                     sampling_freq,
                                                     num_samples)
                wake_index=np.digitize(WakeTimes,time_index)
                Ex[e_inc, wake_index:] = np.dot(np.ones((num_sources)), np.dot(np.ones((num_sinks)), TimeMap[:, :, :(num_samples-wake_index), 0]))
                Ey[e_inc, wake_index:] = np.dot(np.ones((num_sources)), np.dot(np.ones((num_sinks)), TimeMap[:, :, :(num_samples-wake_index), 1]))
                Ez[e_inc, wake_index:] = np.dot(np.ones((num_sources)), np.dot(np.ones((num_sinks)), TimeMap[:, :, :(num_samples-wake_index), 2]))

        else:
            Ex = np.zeros((num_samples), dtype=np.float64)
            Ey = np.zeros((num_samples), dtype=np.float64)
            Ez = np.zeros((num_samples), dtype=np.float64)
            scattering_coefficient = (1 / 4 * scipy.constants.pi)
            TimeMap, WakeTimes = EM.TimeDomainv3(num_sources,
                                                 num_sinks,
                                                 point_informationv2,
                                                 full_index,
                                                 scattering_coefficient,
                                                 wavelength,
                                                 excitation_function,
                                                 sampling_freq,
                                                 num_samples)
            wake_index = np.digitize(WakeTimes, time_index)
            Ex[wake_index:] = np.dot(np.ones((num_sources)),np.dot(np.ones((num_sinks)), TimeMap[:, :, :(num_samples - wake_index), 0]))
            Ey[wake_index:] = np.dot(np.ones((num_sources)),np.dot(np.ones((num_sinks)), TimeMap[:, :, :(num_samples - wake_index), 1]))
            Ez[wake_index:] = np.dot(np.ones((num_sources)),np.dot(np.ones((num_sinks)), TimeMap[:, :, :(num_samples - wake_index), 2]))

        # convert to etheta,ephi

    else:
        # create efiles for model
        if multiE:
            Ex = np.zeros((desired_E_axis.shape[1], num_sinks,num_samples), dtype=np.float64)
            Ey = np.zeros((desired_E_axis.shape[1], num_sinks,num_samples), dtype=np.float64)
            Ez = np.zeros((desired_E_axis.shape[1], num_sinks,num_samples), dtype=np.float64)
            for e_inc in range(desired_E_axis.shape[1]):
                conformal_E_vectors = EM.calculate_conformalVectors(desired_E_axis[e_inc, :],
                                                                    np.asarray(aperture_coords.normals).astype(
                                                                        np.float32))
                for element in range(num_sources):
                    point_informationv2[0:num_sources]['ex'] = 0.0
                    point_informationv2[0:num_sources]['ey'] = 0.0
                    point_informationv2[0:num_sources]['ez'] = 0.0
                    point_informationv2[element]['ex'] = conformal_E_vectors[element, 0] / num_sources
                    point_informationv2[element]['ey'] = conformal_E_vectors[element, 1] / num_sources
                    point_informationv2[element]['ez'] = conformal_E_vectors[element, 2] / num_sources
                    unified_weights[0:num_sources, :] = 0.0
                    unified_weights[element, :] = conformal_E_vectors[element, :] / num_sources
                    scattering_coefficient = (1 / 4 * scipy.constants.pi)
                    TimeMap, WakeTimes = EM.TimeDomainv3(num_sources,
                                                         num_sinks,
                                                         point_informationv2,
                                                         full_index,
                                                         scattering_coefficient,
                                                         wavelength,
                                                         excitation_function,
                                                         sampling_freq,
                                                         num_samples)
                    wake_index = np.digitize(WakeTimes, time_index)
                    Ex[element, wake_index:] = np.dot(np.ones((num_sources)),
                                                      np.dot(np.ones((num_sinks)),
                                                             TimeMap[:, :, :(num_samples - wake_index),0]))
                    Ey[element, wake_index:] = np.dot(np.ones((num_sources)),
                                                      np.dot(np.ones((num_sinks)),
                                                             TimeMap[:, :, :(num_samples - wake_index),1]))
                    Ez[element, wake_index:] = np.dot(np.ones((num_sources)),
                                                      np.dot(np.ones((num_sinks)),
                                                             TimeMap[:, :, :(num_samples - wake_index),2]))

        else:
            Ex = np.zeros((num_sources, num_sinks,num_samples), dtype=np.float64)
            Ey = np.zeros((num_sources, num_sinks,num_samples), dtype=np.float64)
            Ez = np.zeros((num_sources, num_sinks,num_samples), dtype=np.float64)
            for element in range(num_sources):
                point_informationv2[0:num_sources]['ex'] = 0.0
                point_informationv2[0:num_sources]['ey'] = 0.0
                point_informationv2[0:num_sources]['ez'] = 0.0
                point_informationv2[element]['ex'] = conformal_E_vectors[element, 0] / num_sources
                point_informationv2[element]['ey'] = conformal_E_vectors[element, 1] / num_sources
                point_informationv2[element]['ez'] = conformal_E_vectors[element, 2] / num_sources
                unified_weights[0:num_sources, :] = 0.0
                unified_weights[element, :] = conformal_E_vectors[element, :] / num_sources
                scattering_coefficient=(1/4*scipy.constants.pi)
                TimeMap,WakeTimes=EM.TimeDomainv3(num_sources,
                                                  num_sinks,
                                                  point_informationv2,
                                                  full_index,
                                                  scattering_coefficient,
                                                  wavelength,
                                                  excitation_function,
                                                  sampling_freq,
                                                  num_samples)
                Ex[element, :, :] = np.dot(np.ones((num_sources)), np.dot(np.ones((num_sinks)),TimeMap[:, :, :,0]))
                Ey[element, :, :] = np.dot(np.ones((num_sources)), np.dot(np.ones((num_sinks)),TimeMap[:, :, :,1]))
                Ez[element, :, :] = np.dot(np.ones((num_sources)), np.dot(np.ones((num_sinks)),TimeMap[:, :, :,2]))

    return Ex, Ey, Ez, WakeTimes


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy

import numpy as np
import open3d as o3d
import scipy.io as io
from importlib_resources import files

import lyceanem.tests.data
import lyceanem.geometry.targets as tl
import lyceanem.geometry.geometryfunctions as GF


#freq=24e9
#wavelength=3e8/freq
#measurementaddress=Path("./")
#units are mm in the laser scans
#medium_reference='Medium_Reference_Plate_Covered_Normal_1_plan_grid_0p5mm.txt'
#matfile='plate_copper.mat'
#datafile='Medium_Reference_Plate_Covered_Normal_1_plan_grid_0p5mm.txt'
#stream=pkg_resources.resource_stream(__name__,'data/Medium_Reference_Plate_Covered_Normal_1_plan_grid_0p5mm.txt')
#temp=np.loadtxt(stream,delimiter=';',skiprows=2)

#medium_reference_reflector = o3d.geometry.PointCloud()
#medium_reference_reflector.points = o3d.utility.Vector3dVector(temp/1000)
#medium_reference_reflector.estimate_normals()
#reference_point=np.asarray([[np.min(temp[:,0]),np.min(temp[:,1]),np.min(temp[:,2])]])/1000

#downsampled_reflector=medium_reference_reflector.voxel_down_sample(voxel_size=wavelength*0.5)
#downsampled_reflector.translate(-reference_point.ravel(),relative=True)
#downsampled_reflector.translate([-0.15,-0.15,0],relative=True)
def exampleUAV():
    bodystream = files(lyceanem.tests.data).joinpath('UAV.stl')
    arraystream = files(lyceanem.tests.data).joinpath('UAVarray.stl')
    body = o3d.io.read_triangle_mesh(str(bodystream))
    array = o3d.io.read_triangle_mesh(str(arraystream))
    rotation_vector1 = np.asarray([0.0, np.deg2rad(90), 0.0])
    rotation_vector2 = np.asarray([np.deg2rad(90), 0.0, 0.0])
    body=GF.open3drotate(body,o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1))
    body=GF.open3drotate(body,o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector2))
    array=GF.open3drotate(array,o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1))
    array=GF.open3drotate(array,o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector2))
    body.translate(np.asarray([0.25, 0, 0]), relative=True)
    array.translate(np.asarray([0.25, 0, 0]), relative=True)
    array.translate(np.array([-0.18, 0, 0.0125]), relative=True)
    body.translate(np.array([-0.18, 0, 0.0125]), relative=True)
    array.compute_vertex_normals()
    array.paint_uniform_color(np.array([1, 0.259, 0.145]))
    body.compute_vertex_normals()
    body.paint_uniform_color(np.array([0, 0.259, 0.145]))
    return body,array

def prototypescan(mesh_size,wavelength=3e8/24e9,importpcd=True):
    if importpcd:
        if mesh_size<(wavelength*0.5):
            #stream = pkg_resources.resource_stream(__name__,
            #                                       'data/prototypescanpoints24GHztenthwavelength.ply')
            stream=files('lyceanem.tests.data').joinpath('prototypescanpoints24GHztenthwavelength.ply')
            prototype_points=o3d.io.read_point_cloud(str(stream))
        else:
            #stream = pkg_resources.resource_stream(__name__,
            #                                       'data/prototypescanpoints24GHztenthwavelength.ply')
            stream = files(lyceanem.tests.data).joinpath('prototypescanpoints24GHztenthwavelength.ply')
            prototype_points = o3d.io.read_point_cloud(str(stream))
        prototype_mesh,_=tl.meshedReflector(0.3,0.3,6e-3,mesh_size)
    else:
        #stream = pkg_resources.resource_stream(__name__,
        #                                       'data/plate_copper.mat')
        stream = files('lyceanem.tests.data').joinpath('plate_copper.mat')
        matdata=io.loadmat(str(stream))['Flat_prototype']
        copper_reference_point=np.asarray([[np.min(matdata[:,0]),np.min(matdata[:,1]),np.min(matdata[:,2])]])/1000
        prototype= o3d.geometry.PointCloud()
        #convert from mm to si units
        prototype.points= o3d.utility.Vector3dVector(matdata/1000)
        prototype.estimate_normals()
        prototype.orient_normals_to_align_with_direction(orientation_reference=[0., 0., 1.])
        downsampled_prototype=prototype.voxel_down_sample(voxel_size=mesh_size*1.1)
        downsampled_prototype.translate(-copper_reference_point.ravel(),relative=True)
        prototype.translate(-copper_reference_point.ravel(),relative=True)
        downsampled_prototype.translate([-0.15,-0.15,0],relative=True)
        prototype.translate([-0.15,-0.15,0],relative=True)
        reflectorplate,scatter_points=tl.meshedReflector(0.3,0.3,6e-3,mesh_size,sides='all')
        temp_points=np.asarray(scatter_points.points)
        temp_points_back=temp_points[temp_points[:,2]<-1e-3,:]
        temp_points_top=temp_points[temp_points[:,1]>1.49e-1,:]
        top_normals=np.zeros((temp_points_top.shape[0],3))
        top_normals[:,1]=1
        toppoints=o3d.geometry.PointCloud()
        toppoints.points= o3d.utility.Vector3dVector(temp_points_top)
        toppoints.normals= o3d.utility.Vector3dVector(top_normals)
        sideapoints=copy.deepcopy(toppoints)
        rotation_vector1=np.radians(np.array([0,0,90]))
        sideapoints=GF.open3drotate(sideapoints,o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1))
        bottompoints=copy.deepcopy(sideapoints)
        bottompoints=GF.open3drotate(bottompoints,o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1))
        sidebpoints=copy.deepcopy(bottompoints)
        sidebpoints=GF.open3drotate(sidebpoints,o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1))

        back_normals=np.zeros((temp_points_back.shape[0],3))
        back_normals[:,2]=-1
        backpoints=o3d.geometry.PointCloud()
        backpoints.points= o3d.utility.Vector3dVector(temp_points_back)
        backpoints.normals= o3d.utility.Vector3dVector(back_normals)
        #for now only interested in points on the front surface
        temp=np.asarray(downsampled_prototype.points)
        interest_reference=np.min(temp[:,2])
        test_prototype=downsampled_prototype+backpoints+toppoints+bottompoints+sideapoints+sidebpoints
        prototype_mesh,output=o3d.geometry.TriangleMesh.create_from_point_cloud_poisson(test_prototype,linear_fit=True)
        prototype_mesh.compute_triangle_normals()
        prototype_points=o3d.geometry.PointCloud()
        vertices=np.asarray(prototype_mesh.vertices)
        normals=np.asarray(prototype_mesh.vertex_normals)
        prototype_points.points=o3d.utility.Vector3dVector(vertices[vertices[:,2]>=interest_reference,:])
        prototype_points.normals=o3d.utility.Vector3dVector(normals[vertices[:,2]>=interest_reference,:])
    return prototype_mesh,prototype_points

def mediumreferencescan(mesh_size,wavelength=3e8/24e9,importpcd=True):
    if importpcd:
        if mesh_size<(wavelength*0.5):
            #stream = pkg_resources.resource_stream(__name__,
            #                                       'data/referencescanpoints24GHztenthwavelength.ply')
            stream = files('lyceanem.tests.data').joinpath('referencescanpoints24GHztenthwavelength.ply')
            reflector_points=o3d.io.read_point_cloud(str(stream))
        else:
            #stream = pkg_resources.resource_stream(__name__,
            #                                       'data/referencescanpoints24GHztenthwavelength.ply')
            stream = files('lyceanem.tests.data').joinpath('referencescanpoints24GHztenthwavelength.ply')
            reflector_points = o3d.io.read_point_cloud(str(stream))
        reference_mesh,_=tl.meshedReflector(0.3,0.3,6e-3,mesh_size)
    else:
        #stream=pkg_resources.resource_stream(__name__,'data/Medium_Reference_Plate_Covered_Normal_1_plan_grid_0p5mm.txt')
        stream = files('lyceanem.tests.data').joinpath('Medium_Reference_Plate_Covered_Normal_1_plan_grid_0p5mm.txt')
        temp=np.loadtxt(str(stream),delimiter=';',skiprows=2)
        medium_reference_reflector = o3d.geometry.PointCloud()
        medium_reference_reflector.points = o3d.utility.Vector3dVector(temp/1000)
        medium_reference_reflector.estimate_normals()
        medium_reference_reflector.orient_normals_to_align_with_direction(orientation_reference=[0., 0., 1.])
        reference_point=np.asarray([[np.min(temp[:,0]),np.min(temp[:,1]),np.min(temp[:,2])]])/1000
    
        downsampled_reflector=medium_reference_reflector.voxel_down_sample(voxel_size=mesh_size*1.1)
        downsampled_reflector.translate(-reference_point.ravel(),relative=True)
        medium_reference_reflector.translate(-reference_point.ravel(),relative=True)
        downsampled_reflector.translate([-0.15,-0.15,0],relative=True)
        medium_reference_reflector.translate([-0.15,-0.15,0],relative=True)
        reflectorplate,scatter_points=tl.meshedReflector(0.3,0.3,6e-3,mesh_size,sides='all')
        temp_points=np.asarray(scatter_points.points)
        temp_points_back=temp_points[temp_points[:,2]<-1e-3,:]
        temp_points_top=temp_points[temp_points[:,1]>1.49e-1,:]
        top_normals=np.zeros((temp_points_top.shape[0],3))
        top_normals[:,1]=1
        toppoints=o3d.geometry.PointCloud()
        toppoints.points= o3d.utility.Vector3dVector(temp_points_top)
        toppoints.normals= o3d.utility.Vector3dVector(top_normals)
        sideapoints=copy.deepcopy(toppoints)
        rotation_vector1=np.radians(np.array([0,0,90]))
        sideapoints=GF.open3drotate(sideapoints,o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1))

        bottompoints=copy.deepcopy(sideapoints)
        bottompoints=GF.open3drotate(bottompoints,o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1))

        sidebpoints=copy.deepcopy(bottompoints)
        sidebpoints=GF.open3drotate(sidebpoints,o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector1))

        back_normals=np.zeros((temp_points_back.shape[0],3))
        back_normals[:,2]=-1
        backpoints=o3d.geometry.PointCloud()
        backpoints.points= o3d.utility.Vector3dVector(temp_points_back)
        backpoints.normals= o3d.utility.Vector3dVector(back_normals)
        temp=np.asarray(downsampled_reflector.points)
        interest_reference=np.min(temp[:,2])
        test_prototype=downsampled_reflector+backpoints+toppoints+bottompoints+sideapoints+sidebpoints
        reference_mesh,output=o3d.geometry.TriangleMesh.create_from_point_cloud_poisson(test_prototype,linear_fit=True)
        reference_mesh.compute_triangle_normals()
        reflector_points=o3d.geometry.PointCloud()
        vertices=np.asarray(reference_mesh.vertices)
        normals=np.asarray(reference_mesh.vertex_normals)
        reflector_points.points=o3d.utility.Vector3dVector(vertices[vertices[:,2]>=interest_reference,:])
        reflector_points.normals=o3d.utility.Vector3dVector(normals[vertices[:,2]>=interest_reference,:])
    return reference_mesh,reflector_points



import numpy as np
import open3d as o3d
from numba import vectorize
from packaging import version


def tri_areas(solid):
    """
    Takes as an import an open3D triangle mesh, then calculates the area of all triangles in the global units, and
    returns as an array of areas
    """
    triangle_vertices = np.asarray(solid.vertices)
    triangleidx = np.asarray(solid.triangles)
    a_vectors = triangle_vertices[triangleidx[:, 1], :] - triangle_vertices[triangleidx[:, 0], :]
    b_vectors = triangle_vertices[triangleidx[:, 2], :] - triangle_vertices[triangleidx[:, 0], :]
    u = np.cross(a_vectors, b_vectors)
    areas = 0.5 * ((u[:, 0] ** 2 + u[:, 1] ** 2 + u[:, 2] ** 2) ** 0.5)

    return areas


def tri_centroids(solid):
    """
    In order to calculate the centroid of the triangle, take vertices from open3d triangle mesh, then put each triangle
    into origin space, creating oa,ob,oc vectors, then the centroid is a third of the sum of oa+ob+oc, converted back
    into global coordinates
    """
    triangle_vertices = np.asarray(solid.vertices)
    triangleidx = np.asarray(solid.triangles)
    oa = triangle_vertices[triangleidx[:, 0], :]
    ob = triangle_vertices[triangleidx[:, 1], :]
    oc = triangle_vertices[triangleidx[:, 2], :]
    centroids = ((1 / 3) * (oa + ob + oc))
    centroid_cloud = o3d.geometry.PointCloud()
    centroid_cloud.points = o3d.utility.Vector3dVector(centroids)
    centroid_cloud.normals = solid.triangle_normals
    return centroids, centroid_cloud


def decimate_mesh(solid, mesh_sep):
    """
    In order to calculate the scattering appropriately the triangle mesh should be decimated so that the vertices
    are spaced mesh_sep apart.
    inputs are the trianglemesh object solid, and the mesh_sep, and the output is a new solid. This is only required for
    the discrete scattering model, using the centroids or vertices
    """
    new_solid = o3d.geometry.TriangleMesh()
    # lineset=o3d.geometry.LineSet.create_from_triangle_mesh(solid)
    # identify triangles which are too large via areas greater than mesh_sep**2
    area_limit = mesh_sep ** 2
    areas = tri_areas(solid)
    large_tri_index = np.where(areas > area_limit)[0]

    return new_solid


@vectorize(['(float32(float32))', '(float64(float64))'])
def elevationtotheta(el):
    # converting elevation in degrees to theta in degrees
    # elevation is in range -90 to 90 degrees
    # theta is in range 0 to 180 degrees
    if el >= 0.0:
        theta = (90.0 - el)
    else:
        theta = np.abs(el) + 90.0

    return theta


@vectorize(['(float32(float32))', '(float64(float64))'])
def thetatoelevation(theta):
    # converting elevation in degrees to theta in degrees
    # elevation is in range -90 to 90 degrees
    # theta is in range 0 to 180 degrees
    if theta <= 90.0:
        # theta=(90.0-el)
        el = 90 - theta
    else:
        # theta=np.abs(el)+90.0
        el = -(theta - 90.0)

    return el


def open3drotate(item,
                 rotation_matrix,
                 rotation_centre=np.zeros((3, 1), dtype=np.float32)):
    """

    Parameters
    ----------
    item : open3d object
        PointCloud, TriangleMesh or other open3d objects with .rotate function
    rotation_matrix : open3d rotation matrix
        rotation matrix for the desired transformation
    rotation_centre : numpy float (3,1)
        desired rotation centre, defaults to the origin

    Returns
    -------
    item : open3d object
        rotated item
    """
    if version.parse(o3d.__version__) >= version.parse('0.10.0'):
        # new syntax for rotations
        item.rotate(rotation_matrix, center=rotation_centre)
    else:

        item.translate(-1*rotation_centre)
        item.rotate(rotation_matrix, center=False)
        item.translate(rotation_centre)

    return item



import copy
import math
from packaging import version
import numpy as np
from numba import float32, from_dtype, njit, guvectorize
from scipy import interpolate as sp
from scipy.spatial.transform import Rotation as R

import open3d as o3d
from .electromagnetics import beamforming as BM
from .electromagnetics import empropagation as EM
from .geometry import geometryfunctions as GF
from .raycasting import rayfunctions as RF
from glob import glob
import shutil
import os
from sphinx_gallery.scrapers import figure_rst

# A numpy record array (like a struct) to record triangle
point_data = np.dtype([
    # conductivity, permittivity and permiability
    # free space values should be
    # permittivity of free space 8.8541878176e-12F/m
    # permeability of free space 1.25663706212e-6H/m
    ('permittivity', 'c8'), ('permeability', 'c8'),
    # electric or magnetic current sources? E if True
    ('Electric', '?'),
], align=True)
point_t = from_dtype(point_data)  # Create a type that numba can recognize!

# A numpy record array (like a struct) to record triangle
triangle_data = np.dtype([
    # v0 data
    ('v0x', 'f4'), ('v0y', 'f4'), ('v0z', 'f4'),
    # v1 data
    ('v1x', 'f4'), ('v1y', 'f4'), ('v1z', 'f4'),
    # v2 data
    ('v2x', 'f4'), ('v2y', 'f4'), ('v2z', 'f4'),
    # normal vector
    # ('normx', 'f4'),  ('normy', 'f4'), ('normz', 'f4'),
    # ('reflection', np.float64),
    # ('diffuse_c', np.float64),
    # ('specular_c', np.float64),
], align=True)
triangle_t = from_dtype(triangle_data)  # Create a type that numba can recognize!

# ray class, to hold the ray origin, direction, and eventuall other data.
ray_data = np.dtype([
    # origin data
    ('ox', 'f4'), ('oy', 'f4'), ('oz', 'f4'),
    # direction vector
    ('dx', 'f4'), ('dy', 'f4'), ('dz', 'f4'),
    # target
    # direction vector
    # ('tx','f4'),('ty','f4'),('tz','f4'),
    # distance traveled
    ('dist', 'f4'),
    # intersection
    ('intersect', '?'),
], align=True)
ray_t = from_dtype(ray_data)  # Create a type that numba can recognize!
# We can use that type in our device functions and later the kernel!

scattering_point = np.dtype([
    # position data
    ('px', 'f4'), ('py', 'f4'), ('pz', 'f4'),
    # velocity
    ('vx', 'f4'), ('vy', 'f4'), ('vz', 'f4'),
    # normal
    ('nx', 'f4'), ('ny', 'f4'), ('nz', 'f4'),
    # weights
    ('ex', 'c8'), ('ey', 'c8'), ('ez', 'c8'),
    # conductivity, permittivity and permiability
    # free space values should be
    # permittivity of free space 8.8541878176e-12F/m
    # permeability of free space 1.25663706212e-6H/m
    ('permittivity', 'c8'), ('permeability', 'c8'),
    # electric or magnetic current sources? E if True
    ('Electric', '?'),
], align=True)

scattering_t = from_dtype(scattering_point)  # Create a type that numba can recognize!

class PNGScraper(object):
    def __init__(self):
        self.seen = set()

    def __repr__(self):
        return 'PNGScraper'

    def __call__(self, block, block_vars, gallery_conf):
        # Find all PNG files in the directory of this example.
        path_current_example = os.path.dirname(block_vars['src_file'])
        pngs = sorted(glob(os.path.join(path_current_example, '*.png')))

        # Iterate through PNGs, copy them to the sphinx-gallery output directory
        image_names = list()
        image_path_iterator = block_vars['image_path_iterator']
        for png in pngs:
            if png not in self.seen:
                self.seen |= set(png)
                this_image_path = image_path_iterator.next()
                image_names.append(this_image_path)
                shutil.move(png, this_image_path)
        # Use the `figure_rst` helper function to generate rST for image files
        return figure_rst(image_names, gallery_conf['src_dir'])

@njit
def cart2pol(x, y):
    rho = np.sqrt(x ** 2 + y ** 2)
    phi = np.arctan2(y, x)
    return (rho, phi)


@njit
def pol2cart(rho, phi):
    x = rho * np.cos(phi)
    y = rho * np.sin(phi)
    return (x, y)


@njit
def cart2sph(x, y, z):
    # radians
    hxy = np.hypot(x, y)
    r = np.hypot(hxy, z)
    el = np.arctan2(z, hxy)
    az = np.arctan2(y, x)
    return az, el, r


@njit
def sph2cart(az, el, r):
    # radians
    rcos_theta = r * np.cos(el)
    x = rcos_theta * np.cos(az)
    y = rcos_theta * np.sin(az)
    z = r * np.sin(el)
    return x, y, z


@njit
def calc_normals(T):
    # calculate triangle norm
    e1x = T['v1x'] - T['v0x']
    e1y = T['v1y'] - T['v0y']
    e1z = T['v1z'] - T['v0z']

    e2x = T['v2x'] - T['v0x']
    e2y = T['v2y'] - T['v0y']
    e2z = T['v2z'] - T['v0z']

    dirx = e1y * e2z - e1z * e2y
    diry = e1z * e2x - e1x * e2z
    dirz = e1x * e2y - e1y * e2x

    normconst = math.sqrt(dirx ** 2 + diry ** 2 + dirz ** 2)

    T['normx'] = dirx / normconst
    T['normy'] = diry / normconst
    T['normz'] = dirz / normconst

    return T


@guvectorize([(float32[:], float32[:], float32[:], float32)], '(n),(n)->(n),()', target='parallel')
def fast_calc_dv(source, target, dv, normconst):
    dirx = target[0] - source[0]
    diry = target[1] - source[1]
    dirz = target[2] - source[2]
    normconst = math.sqrt(dirx ** 2 + diry ** 2 + dirz ** 2)
    dv = np.array([dirx, diry, dirz]) / normconst


@njit
def calc_dv(source, target):
    dirx = target[0] - source[0]
    diry = target[1] - source[1]
    dirz = target[2] - source[2]
    normconst = np.sqrt(dirx ** 2 + diry ** 2 + dirz ** 2)
    dv = np.array([dirx, diry, dirz]) / normconst
    return dv[0], dv[1], dv[2], normconst


@njit
def calc_dv_norm(source, target, direction, length):
    length[:, 0] = np.sqrt(
        (target[:, 0] - source[:, 0]) ** 2 + (target[:, 1] - source[:, 1]) ** 2 + (target[:, 2] - source[:, 2]) ** 2)
    direction = (target - source) / length
    return direction, length


class structures:
    """
    Structure class to store information about the geometry and materials in the environment, holding the seperate
    shapes as open3D trianglemesh data structures. Everything in the class will be considered an integrated unit, rotating and moving together.
    This class will be developed to include material parameters to enable more complex modelling.

    Units should be SI, metres

    This is the default class for passing structures to the different models.
    """

    def __init__(self,
                 solids):
        # solids is a list of open3D trianglemesh structures
        self.solids = []
        for item in solids:
            self.solids.append(item)
        # self.materials = []
        # for item in material_characteristics:
        #    self.materials.append(item)

    def remove_structure(self, deletion_index):
        """
        removes a component or components from the class
        Parameters
        -----------
        deletion_index : list
            list of integers or numpy array of integers to the solids to be removed

        Returns
        --------
        None
        """
        for entry in range(len(deletion_index)):
            self.solids.pop(deletion_index[entry])
            self.materials.pop(deletion_index[entry])

    def add_structure(self, new_solids):
        """
        adds a component or components from the structure
        Parameters
        -----------
        new_solids : open3d trianglemesh
            the solid to be added to the structure

        Returns
        --------
        None

        """
        self.solids.append(new_solids)
        # self.materials.append(new_materials)

    def rotate_structures(self, rotation_matrix, rotation_centre=np.zeros((3,1), dtype=np.float32)):
        """
        rotates the components of the structure around a common point, default is the origin

        Parameters
        ----------
        rotation_matrix : open3d rotation matrix
            o3d.geometry.TriangleMesh.get_rotation_matrix_from_xyz(rotation_vector)
        rotation_centre : 1*3 numpy float array
            centre of rotation for the structures

        Returns
        --------
        None
        """
        # warning, current commond just rotates around the origin, and until Open3D can be brought up to the
        # latest version without breaking BlueCrystal reqruiements, this will require additional code.
        for item in self.solids:
            self.solids[item]=GF.open3drotate(self.solids[item],rotation_matrix,rotation_centre)


    def translate_structures(self, vector):
        """
        translates the structures in the class by the given cartesian vector (x,y,z)

        Parameters
        -----------
        vector : 1*3 numpy array of floats
            The desired translation vector for the structures

        Returns
        --------
        None
        """
        for item in self.solids:
            self.solids[item].translate(vector)

    def triangles_base_raycaster(self):
        """
        generates the triangles for all the trianglemesh objects in the structure, and outputs them as a continuous array of
        triangle_t format triangles

        Parameters
        -----------
        None

        Returns
        --------
        triangles : N by 1 numpy array of triangle_t triangles
            a continuous array of all the triangles in the structure
        """
        triangles = np.empty((0), dtype=triangle_t)
        for item in self.solids:
            triangles = np.append(triangles, RF.convertTriangles(copy.deepcopy(item)))

        return triangles


class antenna_pattern:
    """
    Antenna Pattern class which allows for patterns to be handled consistently
    across LyceanEM and other modules. The definitions assume that the pattern axes
    are consistent with the global axes set. If a different orientation is required,
    such as a premeasured antenna in a new orientation then the pattern rotate_function
    must be used.

    Antenna Pattern Frequency is in Hz, eventually I may update the class for multi point patterns
    Rotation Offset is Specified in terms of rotations around the x, y, and z axes as roll,pitch/elevation, and azimuth
    in radians.
    """

    def __init__(self,
                 azimuth_resolution=37,
                 elevation_resolution=37,
                 pattern_frequency=1e9,
                 arbitary_pattern=False,
                 arbitary_pattern_type='isotropic',
                 arbitary_pattern_format='Etheta/Ephi',
                 position_mapping=np.zeros((3), dtype=np.float32),
                 rotation_offset=np.zeros((3), dtype=np.float32)):

        self.azimuth_resolution = azimuth_resolution
        self.elevation_resolution = elevation_resolution
        self.pattern_frequency = pattern_frequency
        self.arbitary_pattern_type = arbitary_pattern_type
        self.arbitary_pattern_format = arbitary_pattern_format
        self.position_mapping = position_mapping
        self.rotation_offset = rotation_offset
        self.field_radius=1.0
        az_mesh, elev_mesh = np.meshgrid(np.linspace(-180, 180, self.azimuth_resolution),
                                         np.linspace(-90, 90, self.elevation_resolution))
        self.az_mesh = az_mesh
        self.elev_mesh = elev_mesh
        if self.arbitary_pattern_format == 'Etheta/Ephi':
            self.pattern = np.zeros((self.elevation_resolution, self.azimuth_resolution, 2),
                                    dtype=np.complex64)
        elif self.arbitary_pattern_format == 'ExEyEz':
            self.pattern = np.zeros((self.elevation_resolution, self.azimuth_resolution, 3),
                                    dtype=np.complex64)
        if arbitary_pattern == True:
            self.initilise_pattern()

    def _rotation_matrix(self):
        """_rotation_matrix getter method

        Calculates and returns the (3D) axis rotation matrix.

        Returns
        -------
        : :class:`numpy.ndarray` of shape (3, 3)
            The model (3D) rotation matrix.
        """
        x_rot = R.from_euler('X', self.rotation_offset[0], degrees=True).as_matrix()
        y_rot = R.from_euler('Y', self.rotation_offset[1], degrees=True).as_matrix()
        z_rot = R.from_euler('Z', self.rotation_offset[2], degrees=True).as_matrix()

        # x_rot=np.array([[1,0,0],
        #                [0,np.cos(np.deg2rad(self.rotation_offset[0])),-np.sin(np.deg2rad(self.rotation_offset[0]))],
        #                [0,np.sin(np.deg2rad(self.rotation_offset[0])),np.cos(np.deg2rad(self.rotation_offset[0]))]])

        total_rotation = np.dot(np.dot(z_rot, y_rot), x_rot)
        return total_rotation

    def initilise_pattern(self):
        """
        pattern initialisation function, providing an isotopic pattern
        or quasi-isotropic pattern

        Returns
        -------
        Populated antenna pattern
        """

        if self.arbitary_pattern_type == 'isotropic':
            self.pattern[:, :, 0] = 1.0
        elif self.arbitary_pattern_type == 'xhalfspace':
            az_angles = self.az_mesh[0, :]
            az_index = np.where(np.abs(az_angles) < 90)
            self.pattern[:, az_index, 0] = 1.0
        elif self.arbitary_pattern_type == 'yhalfspace':
            az_angles = self.az_mesh[0, :]
            az_index = np.where(az_angles > 90)
            self.pattern[:, az_index, 0] = 1.0
        elif self.arbitary_pattern_type == 'zhalfspace':
            elev_angles = self.elev_mesh[:, 0]
            elev_index = np.where(elev_angles > 0)
            self.pattern[elev_index, :, 0] = 1.0

    def import_pattern(self, file_location):
        """
        takes the file location and imports the individual pattern file, replacing exsisting values with those of the saved file.

        It is import to note that for CST ASCII export format, that you select a plot range of -180 to 180 for phi, and that by defauly CST exports from 0 to 180 in theta, which
        is the opposite direction to the default for LyceanEM, so this data structures are flipped for consistency.
        Inputs : file location

        Returns : None
        """
        if file_location.suffix == '.txt':
            # file is a CST ffs format
            datafile = np.loadtxt(file_location, skiprows=2)
            theta = datafile[:, 0]
            phi = datafile[:, 1]
            freq_index1 = file_location.name.find('f=')
            freq_index2 = file_location.name.find(')')
            file_frequency = float(file_location.name[freq_index1 + 2:freq_index2]) * 1e9
            phi_steps = np.linspace(np.min(phi), np.max(phi), np.unique(phi).size)
            theta_steps = np.linspace(np.min(theta), np.max(theta), np.unique(theta).size)
            phi_res = np.unique(phi).size
            theta_res = np.unique(theta).size
            etheta = (datafile[:, 3] * np.exp(1j * np.deg2rad(datafile[:, 4]))).reshape(phi_res,theta_res).transpose()
            ephi = (datafile[:, 5] * np.exp(1j * np.deg2rad(datafile[:, 6]))).reshape(phi_res,theta_res).transpose()
            self.azimuth_resolution = phi_res
            self.elevation_resolution = theta_res
            self.elev_mesh = np.flipud(GF.thetatoelevation(theta).reshape(phi_res,theta_res).transpose())
            self.az_mesh = np.flipud(phi.reshape(phi_res,theta_res).transpose())
            self.pattern = np.zeros((self.elevation_resolution, self.azimuth_resolution, 2),
                                    dtype=np.complex64)
            self.pattern[:, :, 0] = np.flipud(etheta)
            self.pattern[:, :, 1] = np.flipud(ephi)
            self.pattern_frequency = file_frequency
        elif file_location.suffix == '.dat':
            # file is .dat format from anechoic chamber measurements
            Ea, Eb, freq, norm, theta_values, phi_values = EM.importDat(file_location)
            az_mesh, elev_mesh = np.meshgrid(phi_values, GF.thetatoelevation(theta_values))
            self.azimuth_resolution = np.unique(phi_values).size
            self.elevation_resolution = np.unique(theta_values).size
            self.az_mesh = az_mesh
            self.elev_mesh = elev_mesh
            self.pattern = np.zeros((self.elevation_resolution, self.azimuth_resolution, 2),
                                    dtype=np.complex64)
            self.pattern_frequency = freq * 1e6
            self.pattern[:, :, 0] = Ea.transpose() + norm
            self.pattern[:, :, 1] = Eb.transpose() + norm

    def export_pattern(self, file_location):
        """
        takes the file location and exports the pattern as a .dat file
        unfinished, must be in Etheta/Ephi format,

        Parameters
        -----------
        file_location : posix path
            the path for the output file, including name

        Returns
        --------
        None
        """
        if self.arbitary_pattern_format == 'ExEyEz':
            self.transmute_pattern()

        theta_flat = GF.elevationtotheta(self.elev_mesh).transpose().reshape(-1,1)
        phi_mesh = self.az_mesh.transpose().reshape(-1,1)
        planes=self.azimuth_resolution
        copolardb=20*np.log10(np.abs(self.pattern[:,:,0].transpose().reshape(-1,1)))
        copolarangle=np.degrees(np.angle(self.pattern[:,:,0].transpose().reshape(-1,1)))
        crosspolardb = 20 * np.log10(np.abs(self.pattern[:, :, 1].transpose().reshape(-1,1)))
        crosspolarangle = np.degrees(np.angle(self.pattern[:, :, 1].transpose().reshape(-1,1)))
        norm=np.nanmax(np.array([np.nanmax(copolardb),np.nanmax(crosspolardb)]))
        copolardb -= norm
        crosspolardb -= norm
        infoarray=np.array([planes,np.min(self.az_mesh),np.max(self.az_mesh),norm,self.pattern_frequency/1e6]).reshape(1,-1)

        dataarray = np.concatenate((theta_flat.reshape(-1,1),
                                    copolardb.reshape(-1,1),
                                    copolarangle.reshape(-1,1),
                                    crosspolardb.reshape(-1,1),
                                    crosspolarangle.reshape(-1,1)),axis=1)
        outputarray=np.concatenate((infoarray,dataarray),axis=0)
        np.savetxt(file_location,outputarray, delimiter=',',fmt="%.2e")


    def display_pattern(self, desired_pattern='both', pattern_min=-40):
        """
        displays the antenna pattern
        """

        if self.arbitary_pattern_format == 'Etheta/Ephi':
            if desired_pattern == 'both':
                BM.PatternPlot(self.pattern[:, :, 0], self.az_mesh, self.elev_mesh, pattern_min=pattern_min,
                               title_text='Etheta')
                BM.PatternPlot(self.pattern[:, :, 1], self.az_mesh, self.elev_mesh, pattern_min=pattern_min,
                               title_text='Ephi')
        elif self.arbitary_pattern_format == 'ExEyEz':
            if desired_pattern == 'both':
                BM.PatternPlot(self.pattern[:, :, 0], self.az_mesh, self.elev_mesh, pattern_min=pattern_min,
                               title_text='Ex')
                BM.PatternPlot(self.pattern[:, :, 1], self.az_mesh, self.elev_mesh, pattern_min=pattern_min,
                               title_text='Ey')
                BM.PatternPlot(self.pattern[:, :, 2], self.az_mesh, self.elev_mesh, pattern_min=pattern_min,
                               title_text='Ez')

    def transmute_pattern(self):
        """
        convert the pattern from Etheta/Ephi format to Ex, Ey,Ez format
        """
        if self.arbitary_pattern_format == 'Etheta/Ephi':
            oldformat = self.pattern.reshape(-1, self.pattern.shape[2])
            old_shape = oldformat.shape
            self.arbitary_pattern_format = 'ExEyEz'
            self.pattern = np.zeros((self.elevation_resolution,
                                     self.azimuth_resolution,
                                     3),
                                    dtype=np.complex64)
            theta = GF.elevationtotheta(self.elev_mesh)
            # convrsion part, move to transmute pattern
            conversion_matrix1 = np.asarray(
                [np.cos(np.deg2rad(theta.ravel())) * np.cos(np.deg2rad(self.az_mesh.ravel())),
                 np.cos(np.deg2rad(theta.ravel())) * np.sin(np.deg2rad(self.az_mesh.ravel())),
                 -np.sin(np.deg2rad(theta.ravel()))]).transpose()
            conversion_matrix2 = np.asarray([-np.sin(np.deg2rad(self.az_mesh.ravel())),
                                             np.cos(np.deg2rad(self.az_mesh.ravel())),
                                             np.zeros(self.az_mesh.size)]).transpose()
            decomposed_fields = oldformat[:, 0].reshape(-1, 1) * conversion_matrix1 + oldformat[:, 1].reshape(-1,
                                                                                                              1) * conversion_matrix2
            self.pattern[:, :, 0] = decomposed_fields[:, 0].reshape(self.elevation_resolution, self.azimuth_resolution)
            self.pattern[:, :, 1] = decomposed_fields[:, 1].reshape(self.elevation_resolution, self.azimuth_resolution)
            self.pattern[:, :, 2] = decomposed_fields[:, 2].reshape(self.elevation_resolution, self.azimuth_resolution)
        else:
            oldformat = self.pattern.reshape(-1, self.pattern.shape[2])
            old_shape = oldformat.shape
            self.arbitary_pattern_format = 'Etheta/Ephi'
            self.pattern = np.zeros((self.elevation_resolution,
                                     self.azimuth_resolution,
                                     2),
                                    dtype=np.complex64)
            theta = GF.elevationtotheta(self.elev_mesh)
            costhetacosphi = (np.cos(np.deg2rad(self.az_mesh.ravel())) * np.cos(np.deg2rad(theta.ravel()))).astype(
                np.complex64)
            sinphicostheta = (np.sin(np.deg2rad(self.az_mesh.ravel())) * np.cos(np.deg2rad(theta.ravel()))).astype(
                np.complex64)
            sintheta = (np.sin(np.deg2rad(theta.ravel()))).astype(np.complex64)
            sinphi = (np.sin(np.deg2rad(self.az_mesh.ravel()))).astype(np.complex64)
            cosphi = (np.cos(np.deg2rad(self.az_mesh.ravel()))).astype(np.complex64)
            new_etheta = oldformat[:, 0] * costhetacosphi + oldformat[:, 1] * sinphicostheta - oldformat[:,
                                                                                               2] * sintheta
            new_ephi = -oldformat[:, 0] * sinphi + oldformat[:, 1] * cosphi
            self.pattern[:, :, 0] = new_etheta.reshape(self.elevation_resolution, self.azimuth_resolution)
            self.pattern[:, :, 1] = new_ephi.reshape(self.elevation_resolution, self.azimuth_resolution)

    def cartesian_points(self):
        """
        exports the cartesian points for all pattern points.
        """
        x, y, z = RF.sph2cart(np.deg2rad(self.az_mesh.ravel()),
                              np.deg2rad(self.elev_mesh.ravel()),
                              np.ones((self.pattern[:, :, 0].size))*self.field_radius)
        field_points = np.array([x, y, z]).transpose().astype(np.float32) + self.position_mapping

        return field_points

    def rotate_pattern(self, rotation_matrix=None):
        """
        Rotate the self pattern from the assumed global axes into the new direction

        Parameters
        ----------
        new_axes : 3x3 numpy float array
            the new vectors for the antenna x,y,z axes

        Returns
        -------
        Updates self.pattern with the new pattern reflecting the antenna
        orientation within the global models

        """
        # generate pattern_coordinates for rotation
        theta = GF.elevationtotheta(self.elev_mesh)
        x, y, z = RF.sph2cart(np.deg2rad(self.az_mesh.ravel()),
                              np.deg2rad(self.elev_mesh.ravel()),
                              np.ones((self.pattern[:, :, 0].size)))
        field_points = np.array([x, y, z]).transpose().astype(np.float32)

        # convert to ExEyEz for rotation
        if self.arbitary_pattern_format == 'Etheta/Ephi':
            self.transmute_pattern()
            desired_format = 'Etheta/Ephi'
        else:
            desired_format = 'ExEyEz'

        if rotation_matrix == None:
            rot_mat = self._rotation_matrix()
        else:
            rot_mat = rotation_matrix

        rotated_points = np.dot(field_points, rot_mat)
        decomposed_fields = self.pattern.reshape(-1, 3)
        # rotation part
        xyzfields = np.dot(decomposed_fields, rot_mat)
        # resample
        resampled_xyzfields = self.resample_pattern(rotated_points, xyzfields, field_points)
        self.pattern[:, :, 0] = resampled_xyzfields[:, 0].reshape(self.elevation_resolution, self.azimuth_resolution)
        self.pattern[:, :, 1] = resampled_xyzfields[:, 1].reshape(self.elevation_resolution, self.azimuth_resolution)
        self.pattern[:, :, 2] = resampled_xyzfields[:, 2].reshape(self.elevation_resolution, self.azimuth_resolution)

        if desired_format == 'Etheta/Ephi':
            # convert back
            self.transmute_pattern()

    def resample_pattern_angular(self, new_azimuth_resolution, new_elevation_resolution):
        """
        resample pattern based upon provided azimuth and elevation resolution
        """
        new_az_mesh, new_elev_mesh = np.meshgrid(np.linspace(-180, 180, new_azimuth_resolution),
                                                 np.linspace(-90, 90, new_elevation_resolution))
        x, y, z = RF.sph2cart(np.deg2rad(new_az_mesh.ravel()),
                              np.deg2rad(new_elev_mesh.ravel()),
                              np.ones((self.pattern[:, :, 0].size)))
        new_field_points = np.array([x, y, z]).transpose().astype(np.float32)
        old_field_points = self.cartesian_points()
        old_pattern = self.pattern.reshape(-1, 2)
        new_pattern = self.resample_pattern(old_field_points, old_pattern, new_field_points)

    def resample_pattern(self, old_points, old_pattern, new_points):
        """

        Parameters
        ----------
        old_points : float xyz
            xyz coordinates that the pattern has been sampled at
        old_pattern : 2 or 3 by n complex array of the antenna pattern at the old_poitns
            DESCRIPTION.
        new_points : desired_grid points in xyz float array
            DESCRIPTION.

        Returns
        -------
        new points, new_pattern

        """
        pol_format = old_pattern.shape[-1]  # will be 2 or three
        new_pattern = np.zeros((new_points.shape[0], pol_format), dtype=np.complex64)
        smoothing_factor = 4
        for component in range(pol_format):
            mag_interpolate = sp.Rbf(old_points[:, 0], old_points[:, 1], old_points[:, 2],
                                     np.abs(old_pattern[:, component]), smooth=smoothing_factor)
            phase_interpolate = sp.Rbf(old_points[:, 0], old_points[:, 1], old_points[:, 2],
                                       np.angle(old_pattern[:, component]), smooth=smoothing_factor)
            new_mag = mag_interpolate(new_points[:, 0], new_points[:, 1], new_points[:, 2])
            new_angles = phase_interpolate(new_points[:, 0], new_points[:, 1], new_points[:, 2])
            new_pattern[:, component] = new_mag * np.exp(1j * new_angles)

        return new_pattern

    def directivity(self):
        """

        Returns
        -------
        Dtheta : numpy array
            directivity for Etheta farfield
        Dphi : numpy array
            directivity for Ephi farfield
        Dtotal : numpy array
            overall directivity pattern
        Dmax : numpy array
            the maximum directivity for each pattern

        """
        Dtheta, Dphi, Dtotal, Dmax = EM.directivity_transformv2(self.pattern[:,:,0],
                                                                self.pattern[:,:,1],
                                                                az_range=self.az_mesh[0,:],
                                                                elev_range=self.elev_mesh[:,0])
        return Dtheta,Dphi,Dtotal,Dmax



import numpy as np
import open3d as o3d

from ..base import scattering_t, calc_dv_norm
from ..electromagnetics import empropagation as EM
from ..geometry import geometryfunctions as GF
from ..geometry import targets as TL
from ..raycasting import rayfunctions as RF


def aperture_projection(aperture,
                        environment=None,
                        wavelength=1.0,
                        az_range=np.linspace(-180.0, 180.0, 19),
                        elev_range=np.linspace(-90.0, 90.0, 19)):
    """Aperture Projection of the provided trianglemesh

    Using the aperture provided and any blocking structures, predict the maximum directivity envelope of the aperture.
    This will initially just cover all the triangles of a provided solid, but eventually I will include a filter list.

    Parameters
    ---------
    aperture : open3D trianglemesh
        trianglemesh of of the desired aperture
    environment : :class: `structures`
        the :class: `structures` class should contain all the environment for scattering, providing the blocking for the rays
    wavelength : float
        the wavelength of interest in metres
    az_range : numpy 1d array of float32
        the azimuth range desired for the farfield pattern in degrees
    elev_range : numpy 1d array of float32
        the elevation range desired for the farfield pattern in degrees

    Returns
    ----------
    directivity_envelope : numpy 2D array of float32
        The predicted maximum directivity envelope for the provided aperture at the wavelength of interest
    pcd : open3d point cloud
        a point cloud colored according to the projected area, normalised to the total projected area of the aperture.
    """
    if environment is None:
        blocking_triangles=RF.convertTriangles(aperture)
    else:
        blocking_triangles = environment.triangles_base_raycaster()

    directivity_envelope = np.zeros((elev_range.shape[0], az_range.shape[0]), dtype=np.float32)
    triangle_centroids,_ = GF.tri_centroids(aperture)
    triangle_areas = GF.tri_areas(aperture)
    triangle_normals = np.asarray(aperture.triangle_normals)
    visible_patterns, pcd = RF.visiblespace(triangle_centroids,
                                          triangle_normals,
                                          blocking_triangles,
                                          vertex_area=triangle_areas,
                                          az_range=az_range,
                                          elev_range=elev_range)
    directivity_envelope[:,:] = (4 * np.pi * visible_patterns) / (wavelength ** 2)

    return directivity_envelope,pcd


def calculate_farfield(aperture_coords,
                       antenna_solid,
                       desired_E_axis,
                       az_range,
                       el_range,
                       scatter_points=None,
                       wavelength=1.0,
                       farfield_distance=2.0,
                       scattering=0,
                       source_weights=None,
                       scattering_weight=1.0,
                       mesh_resolution=0.5,
                       elements=False,
                       los=True,
                       project_vectors=False):
    """
    Based upon the aperture coordinates and solids, predict the farfield for the antenna.

    Parameters
    ---------
    aperture_coords : open3d point cloud
        open3d of the aperture coordinates, from a single point to a mesh sampling across and aperture or surface
    antenna_solid : structure class
        the class should contain all the environment for scattering, providing the blocking for the rays
    desired_E_axis :
        1*3 numpy array of the desired excitation vector
    az_range : 1D numpy array of float
        the desired azimuth planes in degrees
    el_range : 1D numpy array of float
        the desired elevation planes in degrees
    scatter_points : open3d point cloud
        the environment scattering points, defaults to [None]
    wavelength : float
        wavelength of interest in meters, defaults to [1]
    farfield_distance : float
        the distance to evaluate the antenna pattern, defaults to [2]
    scattering: int
     the number of scatters required, if this is set to 0, then only line of sight propagation is considered, defaults to [0]
     source_weights : numpy array complex
        the desired source weights if the source field is to be specified explicitly, should be normalised. This replaces the desired_E_axis specification.
    scattering_weight :
    mesh_resolution :
    elements : boolean
        whether the sources and sinks should be considered as elements of a phased array, or a fixed phase aperture like a horn or reflector
    los : boolean
        The line of sight component can be ignored by setting los to [False], defaults to [True]
    project_vectors : boolean
        should the excitation vector/vectors be projected to be conformal with the surface of the source coordinates

    Returns
    ---------
    etheta : numpy 2D array
        The Etheta farfield component
    ephi : numpy 2D array
        The EPhi farfield component
    """

    # create sink points for the model
    azaz, elel = np.meshgrid(az_range, el_range)
    theta = GF.elevationtotheta(elel)
    sinks = np.zeros((len(np.ravel(azaz)), 3), dtype=np.float32)
    sinks[:, 0], sinks[:, 1], sinks[:, 2] = RF.azeltocart(np.ravel(azaz), np.ravel(elel), farfield_distance)
    sink_normals = np.zeros((len(np.ravel(azaz)), 3), dtype=np.float32)
    origin = np.zeros((len(sinks), 3), dtype=np.float32).ravel()
    lengths = np.zeros((len(np.ravel(azaz)), 1), dtype=np.float32)
    sink_normals, _ = calc_dv_norm(sinks, np.zeros((len(sinks), 3), dtype=np.float32), sink_normals, lengths)
    sink_cloud = RF.points2pointcloud(sinks)
    sink_cloud.normals = o3d.utility.Vector3dVector(sink_normals)
    num_sources = len(np.asarray(aperture_coords.points))
    num_sinks = len(np.asarray(sink_cloud.points))
    environment_triangles = antenna_solid.triangles_base_raycaster()
    if project_vectors:
        conformal_E_vectors = EM.calculate_conformalVectors(desired_E_axis,
                                                            np.asarray(aperture_coords.normals).astype(np.float32))
    else:
        conformal_E_vectors=np.repeat(desired_E_axis.reshape(1,3).astype(np.float32),num_sources,axis=0)
    output_power = 1  # dBwatts
    powerdbm = 10 * np.log10(output_power) + 30
    receiver_impedence=50
    v_transmit = ((10 ** (powerdbm / 20)) * receiver_impedence) ** 0.5
    if scattering == 0:
        # only use the aperture point cloud, no scattering required.
        scatter_points = o3d.geometry.PointCloud()
        unified_model = np.append(np.asarray(aperture_coords.points).astype(np.float32),
                                  np.asarray(sink_cloud.points).astype(np.float32), axis=0)
        unified_normals = np.append(np.asarray(aperture_coords.normals).astype(np.float32),
                                    np.asarray(sink_cloud.normals).astype(np.float32), axis=0)
        unified_weights=np.ones((unified_model.shape[0], 3), dtype=np.complex64)
        if source_weights is None:
            unified_weights[0:num_sources,:] = (conformal_E_vectors / num_sources)*v_transmit  # set total amplitude to 1 for the aperture
        else:
            unified_weights[0:num_sources,:] = source_weights*v_transmit

        unified_weights[num_sources:num_sources + num_sinks,
        :] = 1 #/ num_sinks  # set total amplitude to 1 for the aperture
        point_informationv2 = np.empty((len(unified_model)), dtype=scattering_t)
        # set all sources as magnetic current sources, and permittivity and permeability as free space
        point_informationv2[:]['Electric'] = True
        point_informationv2[:]['permittivity'] = 8.8541878176e-12
        point_informationv2[:]['permeability'] = 1.25663706212e-6
        # set position, velocity, normal, and weight of sources
        point_informationv2[0:num_sources]['px'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['py'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['pz'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 2]
        point_informationv2[0:num_sources]['nx'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['ny'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['nz'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 2]
        # set position and velocity of sinks
        point_informationv2[num_sources:(num_sources + num_sinks)]['px'] = sinks[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['py'] = sinks[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['pz'] = sinks[:, 2]
        point_informationv2[num_sources:(num_sources + num_sinks)]['nx'] = sink_normals[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['ny'] = sink_normals[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['nz'] = sink_normals[:, 2]

        point_informationv2[:]['ex'] = unified_weights[:, 0]
        point_informationv2[:]['ey'] = unified_weights[:, 1]
        point_informationv2[:]['ez'] = unified_weights[:, 2]
        scatter_mask = np.zeros((point_informationv2.shape[0]), dtype=np.int32)
        scatter_mask[0:num_sources] = 0
        scatter_mask[(num_sources + num_sinks):] = 0

    else:
        # create scatter points on antenna solids based upon a half wavelength square
        if scatter_points is None:
            scatter_points, areas = TL.source_cloud_from_shape(antenna_solid, 1e-6,
                                                               (wavelength * mesh_resolution) ** 2 / 2.0)

        unified_model = np.append(np.append(np.asarray(aperture_coords.points).astype(np.float32),
                                            np.asarray(sink_cloud.points).astype(np.float32), axis=0),
                                  np.asarray(scatter_points.points).astype(np.float32), axis=0)
        unified_normals = np.append(np.append(np.asarray(aperture_coords.normals).astype(np.float32),
                                              np.asarray(sink_cloud.normals).astype(np.float32), axis=0),
                                    np.asarray(scatter_points.normals).astype(np.float32), axis=0)
        unified_weights = np.ones((unified_model.shape[0], 3), dtype=np.complex64)
        if source_weights is None:
            unified_weights[0:num_sources, :] = (conformal_E_vectors / num_sources) * v_transmit  # set total amplitude to 1 for the aperture
        else:
            unified_weights[0:num_sources, :] = source_weights * v_transmit
        unified_weights[num_sources:num_sources + num_sinks,
        :] = 1 #/ num_sinks  # set total amplitude to 1 for the aperture
        unified_weights[num_sources + num_sinks:,
        :] = scattering_weight #/ len(np.asarray(scatter_points.points))  # set total amplitude to 1 for the aperture
        point_informationv2 = np.empty((len(unified_model)), dtype=scattering_t)
        # set all sources as magnetic current sources, and permittivity and permeability as free space
        point_informationv2[:]['Electric'] = True
        point_informationv2[:]['permittivity'] = 8.8541878176e-12
        point_informationv2[:]['permeability'] = 1.25663706212e-6
        # set position, velocity, normal, and weight of sources
        point_informationv2[0:num_sources]['px'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['py'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['pz'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 2]
        point_informationv2[0:num_sources]['nx'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['ny'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['nz'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 2]
        # point_informationv2[0:num_sources]['ex']=unified_weights[0:num_sources,0]
        # point_informationv2[0:num_sources]['ey']=unified_weights[0:num_sources,1]
        # point_informationv2[0:num_sources]['ez']=unified_weights[0:num_sources,2]
        # set position and velocity of sinks
        point_informationv2[num_sources:(num_sources + num_sinks)]['px'] = sinks[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['py'] = sinks[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['pz'] = sinks[:, 2]
        # point_informationv2[num_sources:(num_sources+num_sinks)]['vx']=0.0
        # point_informationv2[num_sources:(num_sources+num_sinks)]['vy']=0.0
        # point_informationv2[num_sources:(num_sources+num_sinks)]['vz']=0.0
        point_informationv2[num_sources:(num_sources + num_sinks)]['nx'] = sink_normals[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['ny'] = sink_normals[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['nz'] = sink_normals[:, 2]
        point_informationv2[(num_sources + num_sinks):]['px'] = np.asarray(scatter_points.points).astype(np.float32)[:,
                                                                0]
        point_informationv2[(num_sources + num_sinks):]['py'] = np.asarray(scatter_points.points).astype(np.float32)[:,
                                                                1]
        point_informationv2[(num_sources + num_sinks):]['pz'] = np.asarray(scatter_points.points).astype(np.float32)[:,
                                                                2]
        point_informationv2[(num_sources + num_sinks):]['nx'] = np.asarray(scatter_points.normals).astype(np.float32)[:,
                                                                0]
        point_informationv2[(num_sources + num_sinks):]['ny'] = np.asarray(scatter_points.normals).astype(np.float32)[:,
                                                                1]
        point_informationv2[(num_sources + num_sinks):]['nz'] = np.asarray(scatter_points.normals).astype(np.float32)[:,
                                                                2]
        point_informationv2[:]['ex'] = unified_weights[:, 0]
        point_informationv2[:]['ey'] = unified_weights[:, 1]
        point_informationv2[:]['ez'] = unified_weights[:, 2]
        scatter_mask = np.zeros((point_informationv2.shape[0]), dtype=np.int32)
        scatter_mask[0:num_sources] = 0
        scatter_mask[(num_sources + num_sinks):] = scattering

    # full_index, initial_index = RF.integratedraycastersetup(num_sources,
    #                                                        num_sinks,
    #                                                        point_informationv2,
    #                                                        RF.convertTriangles(antenna_solid),
    #                                                        scatter_mask)
    full_index, rays = RF.workchunkingv2(np.asarray(aperture_coords.points).astype(np.float32),
                                         sinks,
                                         np.asarray(scatter_points.points).astype(np.float32),
                                         environment_triangles,
                                         scattering + 1,
                                         line_of_sight=los)
    print(rays)
    if ~elements:
        # create efiles for model
        etheta = np.zeros((el_range.shape[0], az_range.shape[0]), dtype=np.complex64)
        ephi = np.zeros((el_range.shape[0], az_range.shape[0]), dtype=np.complex64)
        Ex = np.zeros((el_range.shape[0], az_range.shape[0]), dtype=np.complex64)
        Ey = np.zeros((el_range.shape[0], az_range.shape[0]), dtype=np.complex64)
        Ez = np.zeros((el_range.shape[0], az_range.shape[0]), dtype=np.complex64)
        scatter_map = EM.EMGPUFreqDomain(num_sources,
                                         num_sinks,
                                         full_index,
                                         point_informationv2,
                                         wavelength)

        Ex[:, :] = np.sum(scatter_map[:, :, 0], axis=0).reshape(el_range.shape[0], az_range.shape[0])
        Ey[:, :] = np.sum(scatter_map[:, :, 1], axis=0).reshape(el_range.shape[0], az_range.shape[0])
        Ez[:, :] = np.sum(scatter_map[:, :, 2], axis=0).reshape(el_range.shape[0], az_range.shape[0])
        # convert to etheta,ephi
        etheta = Ex * np.cos(np.deg2rad(azaz)) * np.cos(np.deg2rad(theta)) + Ey * np.sin(np.deg2rad(azaz)) * np.cos(
            np.deg2rad(theta)) - Ez * np.sin(np.deg2rad(theta))
        ephi = -Ex * np.sin(np.deg2rad(azaz)) + Ey * np.cos(np.deg2rad(azaz))
    else:
        # create efiles for model
        etheta = np.zeros((num_sources, el_range.shape[0], az_range.shape[0]), dtype=np.complex64)
        ephi = np.zeros((num_sources, el_range.shape[0], az_range.shape[0]), dtype=np.complex64)
        Ex = np.zeros((num_sources, el_range.shape[0], az_range.shape[0]), dtype=np.complex64)
        Ey = np.zeros((num_sources, el_range.shape[0], az_range.shape[0]), dtype=np.complex64)
        Ez = np.zeros((num_sources, el_range.shape[0], az_range.shape[0]), dtype=np.complex64)

        for element in range(num_sources):
            point_informationv2[0:num_sources]['ex'] = 0.0
            point_informationv2[0:num_sources]['ey'] = 0.0
            point_informationv2[0:num_sources]['ez'] = 0.0
            point_informationv2[element]['ex'] = (conformal_E_vectors[element, 0] / num_sources)*v_transmit
            point_informationv2[element]['ey'] = (conformal_E_vectors[element, 1] / num_sources)*v_transmit
            point_informationv2[element]['ez'] = (conformal_E_vectors[element, 2] / num_sources)*v_transmit
            #unified_weights[0:num_sources, :] = 0.0
            #unified_weights[element, :] = (conformal_E_vectors[element, :] / num_sources)*v_transmit
            scatter_map = EM.EMGPUFreqDomain(num_sources,
                                             sinks.shape[0],
                                             full_index,
                                             point_informationv2,
                                             wavelength)
            Ex[element, :, :] = np.dot(np.ones((num_sources)), scatter_map[:, :, 0]).reshape(el_range.shape[0],
                                                                                             az_range.shape[0])
            Ey[element, :, :] = np.dot(np.ones((num_sources)), scatter_map[:, :, 1]).reshape(el_range.shape[0],
                                                                                             az_range.shape[0])
            Ez[element, :, :] = np.dot(np.ones((num_sources)), scatter_map[:, :, 2]).reshape(el_range.shape[0],
                                                                                             az_range.shape[0])
            etheta[element, :, :] = Ex[element, :, :] * np.cos(np.deg2rad(azaz)) * np.cos(np.deg2rad(theta)) + Ey[
                                                                                                               element,
                                                                                                               :,
                                                                                                               :] * np.sin(
                np.deg2rad(azaz)) * np.cos(np.deg2rad(theta)) - Ez[element, :, :] * np.sin(np.deg2rad(theta))
            ephi[element, :, :] = -Ex[element, :, :] * np.sin(np.deg2rad(azaz)) + Ey[element, :, :] * np.cos(
                np.deg2rad(azaz))

    return etheta, ephi


def calculate_scattering(aperture_coords,
                         sink_coords,
                         antenna_solid,
                         desired_E_axis,
                         scatter_points=None,
                         wavelength=1.0,
                         scattering=0,
                         elements=False,
                         los=True,
                         mesh_resolution=0.5,
                         project_vectors=False):
    """
    calculating the scattering from the provided source coordinates, to the provided sink coordinates in the environment.
    This can be used to generate point to point scattering parameters or full scattering networks.

    Parameters
    ----------
    aperture_coords : open3d point cloud
        source coordinates
    sink_coords : open3d point cloud
        sink coordinates
    antenna_solid : structure class
        the class should contain all the environment for scattering, providing the blocking for the rays
    desired_E_axis : 1D numpy array of floats
        the desired excitation vector, can be a 1*3 array or a n*3 array if multiple different exciations are desired in one lauch
    scatter_points : open3d point cloud
        the scattering points in the environment. Defaults to [None], in which case scattering points will be generated from the antenna_solid. If no scattering should be considered then set scattering to [0].
    wavelength : float
        the wavelength of interest in metres
    scattering : int
        the number of reflections to be considered, defaults to [0], but up to 2 can be considered. The higher this number to greater to computational effort, and for most situations 1 should be ample.
    elements : boolean
        whether the sources and sinks should be considered as elements of a phased array, or a fixed phase aperture like a horn or reflector
    los : boolean
        The line of sight component can be ignored by setting los to [False], defaults to [True]
    mesh_resolution : float
        the desired mesh resolution in terms of wavelengths if scattering points are not provided. A scattering mesh is generated on the surfaces of all provided trianglemesh structures.
    project_vectors : boolean


    Returns
    -------
    Ex : numpy array of complex
    Ey : numpy array of complex
    Ez : numpy array of complex

    """
    if desired_E_axis.size > 3:
        # multiple excitations requried
        multiE = True
    else:
        multiE = False

    num_sources = len(np.asarray(aperture_coords.points))
    num_sinks = len(np.asarray(sink_coords.points))
    environment_triangles = antenna_solid.triangles_base_raycaster()
    if scattering == 0:
        # only use the aperture point cloud, no scattering required.
        scatter_points = o3d.geometry.PointCloud()

        if ~multiE:
            if project_vectors:
                conformal_E_vectors = EM.calculate_conformalVectors(desired_E_axis[0,:],
                                                                    np.asarray(aperture_coords.normals).astype(
                                                                        np.float32))
            else:
                conformal_E_vectors = np.repeat(desired_E_axis[0,:].astype(np.float32), num_sources,axis=0).reshape(num_sources,3)
        else:
            if project_vectors:
                conformal_E_vectors = EM.calculate_conformalVectors(desired_E_axis[0,:],
                                                                    np.asarray(aperture_coords.normals).astype(
                                                                        np.float32))
            else:
                conformal_E_vectors = np.repeat(desired_E_axis[0,:].astype(np.float32), num_sources,axis=0).reshape(num_sources,3)

        unified_model = np.append(np.asarray(aperture_coords.points).astype(np.float32),
                                  np.asarray(sink_coords.points).astype(np.float32), axis=0)
        unified_normals = np.append(np.asarray(aperture_coords.normals).astype(np.float32),
                                    np.asarray(sink_coords.normals).astype(np.float32), axis=0)
        unified_weights = np.ones((unified_model.shape[0], 3), dtype=np.complex64)
        unified_weights[0:num_sources,:] = conformal_E_vectors / num_sources  # set total amplitude to 1 for the aperture
        unified_weights[num_sources:num_sources + num_sinks,
        :] = 1 / num_sinks  # set total amplitude to 1 for the aperture
        point_informationv2 = np.empty((len(unified_model)), dtype=scattering_t)
        # set all sources as magnetic current sources, and permittivity and permeability as free space
        point_informationv2[:]['Electric'] = True
        point_informationv2[:]['permittivity'] = 8.8541878176e-12
        point_informationv2[:]['permeability'] = 1.25663706212e-6
        # set position, velocity, normal, and weight of sources
        point_informationv2[0:num_sources]['px'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['py'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['pz'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 2]
        point_informationv2[0:num_sources]['nx'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['ny'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['nz'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 2]
        # set position and velocity of sinks
        point_informationv2[num_sources:(num_sources + num_sinks)]['px'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['py'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['pz'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 2]
        point_informationv2[num_sources:(num_sources + num_sinks)]['nx'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['ny'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['nz'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 2]

        point_informationv2[:]['ex'] = unified_weights[:, 0]
        point_informationv2[:]['ey'] = unified_weights[:, 1]
        point_informationv2[:]['ez'] = unified_weights[:, 2]
        scatter_mask = np.zeros((point_informationv2.shape[0]), dtype=np.int32)
        scatter_mask[0:num_sources] = 0
        scatter_mask[(num_sources + num_sinks):] = 0

    else:
        # create scatter points on antenna solids based upon a half wavelength square
        if scatter_points is None:
            scatter_points, areas = TL.source_cloud_from_shape(antenna_solid, 1e-6, (wavelength * mesh_resolution) ** 2)

        if ~multiE:
            if project_vectors:
                conformal_E_vectors = EM.calculate_conformalVectors(desired_E_axis[0, :],
                                                                    np.asarray(aperture_coords.normals).astype(
                                                                        np.float32))
            else:
                conformal_E_vectors = np.repeat(desired_E_axis[0, :].astype(np.float32), num_sources, axis=0).reshape(num_sources,3)
        else:
            if project_vectors:
                conformal_E_vectors = EM.calculate_conformalVectors(desired_E_axis[0, :],
                                                                    np.asarray(aperture_coords.normals).astype(
                                                                        np.float32))
            else:
                conformal_E_vectors = np.repeat(desired_E_axis[0, :].astype(np.float32), num_sources, axis=0).reshape(num_sources,3)

        unified_model = np.append(np.append(np.asarray(aperture_coords.points).astype(np.float32),
                                            np.asarray(sink_coords.points).astype(np.float32), axis=0),
                                  np.asarray(scatter_points.points).astype(np.float32), axis=0)
        unified_normals = np.append(np.append(np.asarray(aperture_coords.normals).astype(np.float32),
                                              np.asarray(sink_coords.normals).astype(np.float32), axis=0),
                                    np.asarray(scatter_points.normals).astype(np.float32), axis=0)
        unified_weights = np.ones((unified_model.shape[0], 3), dtype=np.complex64)
        unified_weights[0:num_sources,:] = conformal_E_vectors / num_sources  # set total amplitude to 1 for the aperture
        unified_weights[num_sources:num_sources + num_sinks,:] = 1 / num_sinks  # set total amplitude to 1 for the aperture
        unified_weights[num_sources + num_sinks:, :] = 1.0# / len(np.asarray(scatter_points.points))  # set total amplitude to 1 for the aperture
        point_informationv2 = np.empty((len(unified_model)), dtype=scattering_t)
        # set all sources as magnetic current sources, and permittivity and permeability as free space
        point_informationv2[:]['Electric'] = True
        point_informationv2[:]['permittivity'] = 8.8541878176e-12
        point_informationv2[:]['permeability'] = 1.25663706212e-6
        # set position, velocity, normal, and weight of sources
        point_informationv2[0:num_sources]['px'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['py'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['pz'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 2]
        point_informationv2[0:num_sources]['nx'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['ny'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['nz'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 2]
        # point_informationv2[0:num_sources]['ex']=unified_weights[0:num_sources,0]
        # point_informationv2[0:num_sources]['ey']=unified_weights[0:num_sources,1]
        # point_informationv2[0:num_sources]['ez']=unified_weights[0:num_sources,2]
        # set position and velocity of sinks
        point_informationv2[num_sources:(num_sources + num_sinks)]['px'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['py'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['pz'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 2]
        # point_informationv2[num_sources:(num_sources+num_sinks)]['vx']=0.0
        # point_informationv2[num_sources:(num_sources+num_sinks)]['vy']=0.0
        # point_informationv2[num_sources:(num_sources+num_sinks)]['vz']=0.0
        point_informationv2[num_sources:(num_sources + num_sinks)]['nx'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['ny'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['nz'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 2]
        point_informationv2[(num_sources + num_sinks):]['px'] = np.asarray(scatter_points.points).astype(np.float32)[:,
                                                                0]
        point_informationv2[(num_sources + num_sinks):]['py'] = np.asarray(scatter_points.points).astype(np.float32)[:,
                                                                1]
        point_informationv2[(num_sources + num_sinks):]['pz'] = np.asarray(scatter_points.points).astype(np.float32)[:,
                                                                2]
        point_informationv2[(num_sources + num_sinks):]['nx'] = np.asarray(scatter_points.normals).astype(np.float32)[:,
                                                                0]
        point_informationv2[(num_sources + num_sinks):]['ny'] = np.asarray(scatter_points.normals).astype(np.float32)[:,
                                                                1]
        point_informationv2[(num_sources + num_sinks):]['nz'] = np.asarray(scatter_points.normals).astype(np.float32)[:,
                                                                2]
        point_informationv2[:]['ex'] = unified_weights[:, 0]
        point_informationv2[:]['ey'] = unified_weights[:, 1]
        point_informationv2[:]['ez'] = unified_weights[:, 2]
        #scatter_mask = np.zeros((point_informationv2.shape[0]), dtype=np.int32)
        #scatter_mask[0:num_sources] = 0
        #scatter_mask[(num_sources + num_sinks):] = scattering

    # full_index, initial_index = RF.integratedraycastersetup(num_sources,
    #                                                        num_sinks,
    #                                                        point_informationv2,
    #                                                        RF.convertTriangles(antenna_solid),
    #                                                        scatter_mask)

    full_index, rays = RF.workchunkingv2(np.asarray(aperture_coords.points).astype(np.float32),
                                         np.asarray(sink_coords.points).astype(np.float32),
                                         np.asarray(scatter_points.points).astype(np.float32),
                                         environment_triangles, scattering + 1,
                                         line_of_sight=los)


    if (not elements):
        # create efiles for model
        if multiE:
            Ex = np.zeros((desired_E_axis.shape[0]), dtype=np.complex64)
            Ey = np.zeros((desired_E_axis.shape[0]), dtype=np.complex64)
            Ez = np.zeros((desired_E_axis.shape[0]), dtype=np.complex64)
            for e_inc in range(desired_E_axis.shape[0]):
                conformal_E_vectors = EM.calculate_conformalVectors(desired_E_axis[e_inc, :],
                                                                    np.asarray(aperture_coords.normals).astype(
                                                                        np.float32))
                unified_weights[0:num_sources, :] = conformal_E_vectors / num_sources
                point_informationv2[:]['ex'] = unified_weights[:, 0]
                point_informationv2[:]['ey'] = unified_weights[:, 1]
                point_informationv2[:]['ez'] = unified_weights[:, 2]
                scatter_map = EM.EMGPUFreqDomain(num_sources,
                                                 num_sinks,
                                                 full_index,
                                                 point_informationv2,
                                                 wavelength)
                Ex[e_inc] = np.sum(np.dot(np.ones((num_sources)), scatter_map[:, :, 0]))
                Ey[e_inc] = np.sum(np.dot(np.ones((num_sources)), scatter_map[:, :, 1]))
                Ez[e_inc] = np.sum(np.dot(np.ones((num_sources)), scatter_map[:, :, 2]))
        else:
            scatter_map = EM.EMGPUFreqDomain(num_sources,
                                             num_sinks,
                                             full_index,
                                             point_informationv2,
                                             wavelength)
            Ex = np.sum(np.dot(np.ones((num_sources)), scatter_map[:, :, 0]))
            Ey = np.sum(np.dot(np.ones((num_sources)), scatter_map[:, :, 1]))
            Ez = np.sum(np.dot(np.ones((num_sources)), scatter_map[:, :, 2]))

        # convert to etheta,ephi

    else:
        # create efiles for model
        if multiE:
            Ex = np.zeros((desired_E_axis.shape[1], num_sinks), dtype=np.complex64)
            Ey = np.zeros((desired_E_axis.shape[1], num_sinks), dtype=np.complex64)
            Ez = np.zeros((desired_E_axis.shape[1], num_sinks), dtype=np.complex64)
            for e_inc in range(desired_E_axis.shape[1]):
                conformal_E_vectors = EM.calculate_conformalVectors(desired_E_axis[e_inc, :],
                                                                    np.asarray(aperture_coords.normals).astype(
                                                                        np.float32))
                for element in range(num_sources):
                    point_informationv2[0:num_sources]['ex'] = 0.0
                    point_informationv2[0:num_sources]['ey'] = 0.0
                    point_informationv2[0:num_sources]['ez'] = 0.0
                    point_informationv2[element]['ex'] = conformal_E_vectors[element, 0] / num_sources
                    point_informationv2[element]['ey'] = conformal_E_vectors[element, 1] / num_sources
                    point_informationv2[element]['ez'] = conformal_E_vectors[element, 2] / num_sources
                    unified_weights[0:num_sources, :] = 0.0
                    unified_weights[element, :] = conformal_E_vectors[element, :] / num_sources
                    scatter_map = EM.EMGPUFreqDomain(num_sources,
                                                     num_sinks,
                                                     full_index,
                                                     point_informationv2,
                                                     wavelength)
                    Ex[element, :, e_inc] = np.dot(np.ones((num_sources)), scatter_map[:, :, 0])
                    Ey[element, :, e_inc] = np.dot(np.ones((num_sources)), scatter_map[:, :, 1])
                    Ez[element, :, e_inc] = np.dot(np.ones((num_sources)), scatter_map[:, :, 2])
        else:
            Ex = np.zeros((num_sources, num_sinks), dtype=np.complex64)
            Ey = np.zeros((num_sources, num_sinks), dtype=np.complex64)
            Ez = np.zeros((num_sources, num_sinks), dtype=np.complex64)
            for element in range(num_sources):
                point_informationv2[0:num_sources]['ex'] = 0.0
                point_informationv2[0:num_sources]['ey'] = 0.0
                point_informationv2[0:num_sources]['ez'] = 0.0
                point_informationv2[element]['ex'] = conformal_E_vectors[element, 0] / num_sources
                point_informationv2[element]['ey'] = conformal_E_vectors[element, 1] / num_sources
                point_informationv2[element]['ez'] = conformal_E_vectors[element, 2] / num_sources
                unified_weights[0:num_sources, :] = 0.0
                unified_weights[element, :] = conformal_E_vectors[element, :] / num_sources
                scatter_map = EM.EMGPUFreqDomain(num_sources,
                                                 num_sinks,
                                                 full_index,
                                                 point_informationv2,
                                                 wavelength)
                Ex[element, :] = np.dot(np.ones((num_sources)), scatter_map[:, :, 0])
                Ey[element, :] = np.dot(np.ones((num_sources)), scatter_map[:, :, 1])
                Ez[element, :] = np.dot(np.ones((num_sources)), scatter_map[:, :, 2])

    return Ex, Ey, Ez

def calculate_scattering_isotropic(aperture_coords,
                                   sink_coords,
                                   antenna_solid,
                                   scatter_points=None,
                                   wavelength=1.0,
                                   scattering=0,
                                   elements=False,
                                   mesh_resolution=0.5):
    """
    Based upon the aperture coordinates and solids, predict the farfield for the antenna.
    """
    num_sources = len(np.asarray(aperture_coords.points))
    num_sinks = len(np.asarray(sink_coords.points))

    if scattering == 0:
        # only use the aperture point cloud, no scattering required.
        scatter_points = o3d.geometry.PointCloud()

        unified_model = np.append(np.asarray(aperture_coords.points).astype(np.float32),
                                  np.asarray(sink_coords.points).astype(np.float32), axis=0)
        unified_normals = np.append(np.asarray(aperture_coords.normals).astype(np.float32),
                                    np.asarray(sink_coords.normals).astype(np.float32), axis=0)
        unified_weights = np.ones((unified_model.shape[0], 3), dtype=np.complex64)
        unified_weights[0:num_sources,:] = 1.0 / num_sources  # set total amplitude to 1 for the aperture
        unified_weights[num_sources:num_sources + num_sinks,
        :] = 1 / num_sinks  # set total amplitude to 1 for the aperture
        point_informationv2 = np.empty((len(unified_model)), dtype=scattering_t)
        # set all sources as magnetic current sources, and permittivity and permeability as free space
        point_informationv2[:]['Electric'] = True
        point_informationv2[:]['permittivity'] = 8.8541878176e-12
        point_informationv2[:]['permeability'] = 1.25663706212e-6
        # set position, velocity, normal, and weight of sources
        point_informationv2[0:num_sources]['px'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['py'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['pz'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 2]
        point_informationv2[0:num_sources]['nx'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['ny'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['nz'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 2]
        # set position and velocity of sinks
        point_informationv2[num_sources:(num_sources + num_sinks)]['px'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['py'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['pz'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 2]
        point_informationv2[num_sources:(num_sources + num_sinks)]['nx'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['ny'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['nz'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 2]

        point_informationv2[:]['ex'] = unified_weights[:, 0]
        point_informationv2[:]['ey'] = unified_weights[:, 1]
        point_informationv2[:]['ez'] = unified_weights[:, 2]
        scatter_mask = np.zeros((point_informationv2.shape[0]), dtype=np.int32)
        scatter_mask[0:num_sources] = 0
        scatter_mask[(num_sources + num_sinks):] = 0

    else:
        # create scatter points on antenna solids based upon a half wavelength square
        if scatter_points is None:
            scatter_points, areas = TL.source_cloud_from_shape(antenna_solid, 1e-6, (wavelength * mesh_resolution) ** 2)

        unified_model = np.append(np.append(np.asarray(aperture_coords.points).astype(np.float32),
                                            np.asarray(sink_coords.points).astype(np.float32), axis=0),
                                  np.asarray(scatter_points.points).astype(np.float32), axis=0)
        unified_normals = np.append(np.append(np.asarray(aperture_coords.normals).astype(np.float32),
                                              np.asarray(sink_coords.normals).astype(np.float32), axis=0),
                                    np.asarray(scatter_points.normals).astype(np.float32), axis=0)
        unified_weights = np.ones((unified_model.shape[0], 3), dtype=np.complex64)
        unified_weights[0:num_sources,:] = 1.0 / num_sources  # set total amplitude to 1 for the aperture
        unified_weights[num_sources:num_sources + num_sinks,:] = 1 / num_sinks  # set total amplitude to 1 for the aperture
        unified_weights[num_sources + num_sinks:, :] = 1.0# / len(np.asarray(scatter_points.points))  # set total amplitude to 1 for the aperture
        point_informationv2 = np.empty((len(unified_model)), dtype=scattering_t)
        # set all sources as magnetic current sources, and permittivity and permeability as free space
        point_informationv2[:]['Electric'] = True
        point_informationv2[:]['permittivity'] = 8.8541878176e-12
        point_informationv2[:]['permeability'] = 1.25663706212e-6
        # set position, velocity, normal, and weight of sources
        point_informationv2[0:num_sources]['px'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['py'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['pz'] = np.asarray(aperture_coords.points).astype(np.float32)[:, 2]
        point_informationv2[0:num_sources]['nx'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 0]
        point_informationv2[0:num_sources]['ny'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 1]
        point_informationv2[0:num_sources]['nz'] = np.asarray(aperture_coords.normals).astype(np.float32)[:, 2]
        # point_informationv2[0:num_sources]['ex']=unified_weights[0:num_sources,0]
        # point_informationv2[0:num_sources]['ey']=unified_weights[0:num_sources,1]
        # point_informationv2[0:num_sources]['ez']=unified_weights[0:num_sources,2]
        # set position and velocity of sinks
        point_informationv2[num_sources:(num_sources + num_sinks)]['px'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['py'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['pz'] = np.asarray(sink_coords.points).astype(
            np.float32)[:, 2]
        # point_informationv2[num_sources:(num_sources+num_sinks)]['vx']=0.0
        # point_informationv2[num_sources:(num_sources+num_sinks)]['vy']=0.0
        # point_informationv2[num_sources:(num_sources+num_sinks)]['vz']=0.0
        point_informationv2[num_sources:(num_sources + num_sinks)]['nx'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 0]
        point_informationv2[num_sources:(num_sources + num_sinks)]['ny'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 1]
        point_informationv2[num_sources:(num_sources + num_sinks)]['nz'] = np.asarray(sink_coords.normals).astype(
            np.float32)[:, 2]
        point_informationv2[(num_sources + num_sinks):]['px'] = np.asarray(scatter_points.points).astype(np.float32)[:,
                                                                0]
        point_informationv2[(num_sources + num_sinks):]['py'] = np.asarray(scatter_points.points).astype(np.float32)[:,
                                                                1]
        point_informationv2[(num_sources + num_sinks):]['pz'] = np.asarray(scatter_points.points).astype(np.float32)[:,
                                                                2]
        point_informationv2[(num_sources + num_sinks):]['nx'] = np.asarray(scatter_points.normals).astype(np.float32)[:,
                                                                0]
        point_informationv2[(num_sources + num_sinks):]['ny'] = np.asarray(scatter_points.normals).astype(np.float32)[:,
                                                                1]
        point_informationv2[(num_sources + num_sinks):]['nz'] = np.asarray(scatter_points.normals).astype(np.float32)[:,
                                                                2]
        point_informationv2[:]['ex'] = unified_weights[:, 0]
        point_informationv2[:]['ey'] = unified_weights[:, 1]
        point_informationv2[:]['ez'] = unified_weights[:, 2]
        scatter_mask = np.zeros((point_informationv2.shape[0]), dtype=np.int32)
        scatter_mask[0:num_sources] = 0
        scatter_mask[(num_sources + num_sinks):] = scattering

    # full_index, initial_index = RF.integratedraycastersetup(num_sources,
    #                                                        num_sinks,
    #                                                        point_informationv2,
    #                                                        RF.convertTriangles(antenna_solid),
    #                                                        scatter_mask)
    full_index, rays = RF.workchunkingv2(np.asarray(aperture_coords.points).astype(np.float32),
                                         np.asarray(sink_coords.points).astype(np.float32),
                                         np.asarray(scatter_points.points).astype(np.float32),
                                         RF.convertTriangles(antenna_solid), scattering + 1)

    if (not elements):
        # create efiles for model
            scatter_map = EM.IsoGPUFreqDomain(num_sources,
                                              num_sinks,
                                              full_index,
                                              point_informationv2,
                                              wavelength)
            final_scattering_map = np.sum(np.dot(np.ones((num_sources)), scatter_map[:, :, 0]))

        # convert to etheta,ephi

    else:
        # create efiles for model

        final_scattering_map = np.zeros((num_sources, num_sinks), dtype=np.complex64)

        for element in range(num_sources):
            point_informationv2[0:num_sources]['ex'] = 0.0
            point_informationv2[0:num_sources]['ey'] = 0.0
            point_informationv2[0:num_sources]['ez'] = 0.0
            point_informationv2[element]['ex'] = 1.0 / num_sources
            point_informationv2[element]['ey'] = 1.0 / num_sources
            point_informationv2[element]['ez'] = 1.0 / num_sources
            unified_weights[0:num_sources, :] = 0.0
            unified_weights[element, :] = 1.0 / num_sources
            scatter_map = EM.EMGPUFreqDomain(num_sources,
                                             num_sinks,
                                             full_index,
                                             point_informationv2,
                                             wavelength)
            final_scattering_map[element, :] = np.dot(np.ones((num_sources)), scatter_map[:, :, 0])

    return final_scattering_map